/* Copyright (c) 1998 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/* granular sampling of a table*/



#include "m_imp.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#ifdef NT
#pragma warning (disable: 4305 4244)
#endif


EXTERN void garray_usedindsp(t_garray *x);

/* max so far, 27 */
#define DEFNSTRMS	4
#define MAXSTRMS	50
#define	DEFGLEN		4410


/******************** grantabr~ ***********************/

static t_class *grantabr_tilde_class;

typedef struct _stream
{
	int st_pos;		/* position of the grain in the input array */
	float *st_ix;	/* index into the array: point from which to read */
	int st_phase;	/* current grain phase, used to calc ix and envelope */
	char st_free;	/* stream free after a grain flag */
	char st_trigd;	/* stream armed for triggering a grain */
	char st_on;		/* stream switched on */
	int st_glen;	/* actual length of a grain in this stream */
} t_stream;

typedef struct _grantabr
{
    t_object x_obj;	/* this object */
    int x_npoints;	/* size of input array */
    float *x_vec;	/* start of the array */
    t_symbol *x_arrayname;	/* name of the array*/
	t_stream x_stream[MAXSTRMS];	/* array of granular streams */
	int x_gin;	/* length of grain from control input */
	float x_dens;	/* density of grains in a stream */
	int x_nsin;	/* number of streams from control input */
	int x_nsact;	/* actual current number of streams */
} t_grantabr;




static t_int *grantabr_tilde_perform(t_int *w)
{
	t_grantabr *x = (t_grantabr *)(w[1]);
	t_float *out = (t_float *)(w[2]);
	register int n = w[3];
	register int j, k;
	float sum;
//	float M_PI = 3.14159265358979;

/* so it doesn't get stuck when number of streams falls to zero, keep checking:
if no. is zero, but the control input goes up, then activate the first stream.
This can be done every block rather than every sample */
	if ((!x->x_nsact)&&(x->x_nsin)) {
		x->x_stream[0].st_on = 1;
		x->x_nsact++;
	}



	while (n--){

/* go through all the streams possible one by one */
		j = MAXSTRMS;
		sum = 0;

		while (j--){

/* if the stream's active, do it's stuff.. */
			if (x->x_stream[j].st_on) {


/* if stream's free, trigger a grain*/
				if ((x->x_stream[j].st_free==1) && (!x->x_stream[j].st_trigd)){

					if ((rand()/(float)RAND_MAX) > x->x_dens){
						x->x_stream[j].st_trigd = 1;
						x->x_stream[j].st_pos = (rand()/(float)RAND_MAX) * (x->x_npoints-x->x_stream[j].st_glen);
				   		x->x_stream[j].st_free = 0;
					}

				}



				if (x->x_stream[j].st_trigd){

/* calculate the index into the array */
					x->x_stream[j].st_ix =  x->x_vec + x->x_stream[j].st_pos + x->x_stream[j].st_phase++;

/* if the grain ends, free the stream and untrigger it; */
 	 				if((x->x_stream[j].st_free==0) && (x->x_stream[j].st_phase >= x->x_stream[j].st_glen)){
						x->x_stream[j].st_phase = 0;
			   			x->x_stream[j].st_free = 1;
						x->x_stream[j].st_trigd = 0;

/* also update the number of streams if there's any change...
if there's a decrease, turn off this stream */
						if (x->x_nsin < x->x_nsact){
							x->x_stream[j].st_on = 0;
							x->x_nsact--;
						}
/* otherwise if there's an increase, find the next available switched off stream, switch it on
and update it's grain length to the current value...ready for next time round */
						else {
							k = MAXSTRMS;
							while ((x->x_nsin > x->x_nsact)&&(k--)) {
								if (!x->x_stream[k].st_on){
									x->x_stream[k].st_on = 1;
									x->x_stream[k].st_glen = x->x_gin;
									x->x_nsact++;
								}
							}
						}
/* and update the grain length for this stream */
						x->x_stream[j].st_glen = x->x_gin;
 					}
 				}
 			}
/* envelope the grain... */
			if (x->x_stream[j].st_trigd) sum = sum + (*x->x_stream[j].st_ix)*
				((1+cos(M_PI*(1+(2*(float)(x->x_stream[j].st_phase-1)/(float)x->x_stream[j].st_glen))))/2);
		}

		*out++ = sum;

	}
	return (w+4);
}



void grantabr_tilde_set(t_grantabr *x, t_symbol *s)
{
    t_garray *a;
	int n;

    x->x_arrayname = s;
    if (!(a = (t_garray *)pd_findbyclass(x->x_arrayname, garray_class)))
    {
    	if (*s->s_name)
    	    error("grantabr~: %s: no such array", x->x_arrayname->s_name);
    	x->x_vec = 0;
    }
    else if (!garray_getfloatarray(a, &x->x_npoints, &x->x_vec))
    {
    	error("%s: bad template for grantabr~", x->x_arrayname->s_name);
    	x->x_vec = 0;
    }
    else garray_usedindsp(a);

	for (n=0; n<MAXSTRMS; n++){
		x->x_stream[n].st_ix = x->x_vec;
	}
}

void grantabr_tilde_grainlen(t_grantabr *x, t_floatarg f)
{
	if ((f>=0)&&(f<x->x_npoints)) x->x_gin = (int)f;
}

void grantabr_tilde_nstreams(t_grantabr *x, t_floatarg f)
{
	if ((f>=0)&&(f<=MAXSTRMS)) x->x_nsin = (int)f;
}

void grantabr_tilde_density(t_grantabr *x, t_floatarg f)
{
	if ((f>0)&&(f<1)) x->x_dens = (float)(1 - (f/1000));
}

static void grantabr_tilde_dsp(t_grantabr *x, t_signal **sp)
{
    grantabr_tilde_set(x, x->x_arrayname);

    dsp_add(grantabr_tilde_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
}

static void grantabr_tilde_free(t_grantabr *x)
{
}

static void *grantabr_tilde_new(t_symbol *s)
{
	int n;

    t_grantabr *x = (t_grantabr *)pd_new(grantabr_tilde_class);
    x->x_arrayname = s;

	for (n=0; n<MAXSTRMS; n++){
		x->x_stream[n].st_free = 1;
		x->x_stream[n].st_phase = 0;
		x->x_stream[n].st_trigd = 0;
		x->x_stream[n].st_glen = DEFGLEN;
		x->x_stream[n].st_on = 0;
	}

	x->x_nsact = DEFNSTRMS;

	for (n=0; n<x->x_nsact; n++){
		x->x_stream[n].st_on = 1;
	}

	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("densin"));
 	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("glenin"));
	outlet_new(&x->x_obj, &s_signal);
    return (x);
}

void grantabr_tilde_setup(void)
{
    grantabr_tilde_class = class_new(gensym("grantabr~"),
    	(t_newmethod)grantabr_tilde_new, (t_method)grantabr_tilde_free,
    	sizeof(t_grantabr), 0, A_DEFSYM, 0);

    class_addmethod(grantabr_tilde_class, nullfn, gensym("signal"), 0);
    class_addmethod(grantabr_tilde_class, (t_method)grantabr_tilde_dsp,
    	gensym("dsp"), 0);
    class_addmethod(grantabr_tilde_class, (t_method)grantabr_tilde_set,
    	gensym("set"), A_SYMBOL, 0);
	class_addmethod(grantabr_tilde_class, (t_method)grantabr_tilde_nstreams,
    	gensym("float"), A_FLOAT, 0);
	class_addmethod(grantabr_tilde_class, (t_method)grantabr_tilde_grainlen,
    	gensym("glenin"), A_FLOAT, 0);
	class_addmethod(grantabr_tilde_class, (t_method)grantabr_tilde_density,
    	gensym("densin"), A_FLOAT, 0);
}



/* ------------------------ global setup routine ------------------------- */

void grantabr_setup(void)
{
	grantabr_tilde_setup();
}
