/* Copyright (c) 1998 The Regents of the University of California.
* For information on usage and redistribution, and for a DISCLAIMER OF ALL
* WARRANTIES, see the file, "LICENSE.txt," in this distribution.  */

/* continuous sampling into variable-size table, for gransamp */


#include "m_imp.h"

#ifdef NT
#pragma warning (disable: 4305 4244)
#endif


EXTERN void garray_usedindsp(t_garray *x);

/* ------------------------- grantabw~ -------------------------- */

static t_class *grantabw_tilde_class;

typedef struct _grantabw_tilde
{
    t_object x_obj;
    int x_phase;
    int x_nsampsintab;
	int x_nsampstowrite;
    float *x_vec;
    t_symbol *x_arrayname;
    t_clock *x_clock;
} t_grantabw_tilde;

static t_int *grantabw_tilde_perform(t_int *w)
{
    t_grantabw_tilde *x = (t_grantabw_tilde *)(w[1]);
    t_float *in = (t_float *)(w[2]);
    int n = (int)(w[3]), phase = x->x_phase, endphase = x->x_nsampstowrite;
    if (!x->x_vec) goto bad;

    if (endphase > phase)
    {
    	int nxfer = endphase - phase;
    	float *fp = x->x_vec + phase;
    	if (nxfer > n) nxfer = n;
    	phase += nxfer;
    	while (nxfer--) *fp++ = *in++;
    	if (phase >= endphase)
    	{
    	    clock_delay(x->x_clock, 0);
    	    phase = 0x7fffffff;
    	}
    	x->x_phase = phase;
    }
	else x->x_phase = phase = 0;
	
bad:
    return (w+4);
}

void grantabw_tilde_set(t_grantabw_tilde *x, t_symbol *s)
{
    t_garray *a;

    x->x_arrayname = s;
    if (!(a = (t_garray *)pd_findbyclass(x->x_arrayname, garray_class)))
    {
    	if (*s->s_name) pd_error(x, "grantabw~: %s: no such array",
    	    x->x_arrayname->s_name);
    	x->x_vec = 0;
    }
    else if (!garray_getfloatarray(a, &x->x_nsampsintab, &x->x_vec))
    {
    	error("%s: bad template for tabread4~", x->x_arrayname->s_name);
    	x->x_vec = 0;
    }
    else garray_usedindsp(a);
}

static void grantabw_tilde_dsp(t_grantabw_tilde *x, t_signal **sp)
{
    grantabw_tilde_set(x, x->x_arrayname);
    dsp_add(grantabw_tilde_perform, 3, x, sp[0]->s_vec, sp[0]->s_n);
}


/* end of writable bit of table */
static void grantabw_tilde_float(t_grantabw_tilde *x, t_floatarg f)
{
	if (f > x->x_nsampsintab) f = x->x_nsampsintab;
	if (f < 0) f = 0;
    x->x_nsampstowrite = f;
}

static void grantabw_tilde_tick(t_grantabw_tilde *x)
{
    t_garray *a = (t_garray *)pd_findbyclass(x->x_arrayname, garray_class);
    if (!a) bug("grantabw_tilde_tick");
    else garray_redraw(a);
}

static void grantabw_tilde_free(t_grantabw_tilde *x)
{
    clock_free(x->x_clock);
}

static void *grantabw_tilde_new(t_symbol *s)
{
    t_grantabw_tilde *x = (t_grantabw_tilde *)pd_new(grantabw_tilde_class);
    x->x_clock = clock_new(x, (t_method)grantabw_tilde_tick);
    x->x_phase = 0x7fffffff;
    x->x_arrayname = s;
	inlet_new(&x->x_obj, &x->x_obj.ob_pd, gensym("float"), gensym("ft1"));
    return (x);
}

void grantabw_tilde_setup(void)
{
    grantabw_tilde_class = class_new(gensym("grantabw~"),
    	(t_newmethod)grantabw_tilde_new, (t_method)grantabw_tilde_free,
    	sizeof(t_grantabw_tilde), 0, A_DEFSYM, 0);
    class_addmethod(grantabw_tilde_class, nullfn, gensym("signal"), 0);
    class_addmethod(grantabw_tilde_class, (t_method)grantabw_tilde_dsp,
    	gensym("dsp"), 0);
    class_addmethod(grantabw_tilde_class, (t_method)grantabw_tilde_set,
    	gensym("set"), A_SYMBOL, 0);
	class_addmethod(grantabw_tilde_class, (t_method)grantabw_tilde_float,
gensym("ft1"), A_FLOAT, 0); }


/* ------------------------ global setup routine ------------------------- */

void grantabw_setup(void)
{
    grantabw_tilde_setup();
}

